import redis
import time


class RedisQueue(object):
    """
        A FIFO queue based on Redis. Can be used by multiple workers.
        Workers can share the queue by specifying the same queue name.

        Args:
            queue (str): the name of a queue.
            unique_history (bool): Any element that is ever pushed into the queue is not pushed again.
            soft_capacity (int): The max size of queue. This is 'soft' because the queue can grow up to `soft_capacity+(number of workers)-1`.
            redis_client: if not provided, the default one is created.
    """

    def __init__(self, queue: str, unique_history: bool = False, soft_capacity=None, redis_client=None):
        self.queue = queue
        self.unique_history = unique_history
        if self.unique_history:
            self.history_queue = self.queue + '_all'
        if redis_client == None:
            redis_client = redis.StrictRedis(host='localhost', port=6379, db=0, decode_responses=True)
        self.soft_capacity = soft_capacity
        self.cooldown = 0.1
        self.client = redis_client

    def clear(self, history: bool = False):
        """
        Empty the queue.

        Args:
            history: if True, clear all history as well. has no effect if `unique_history` is False.
        """
        self.client.delete(self.queue)
        if history and self.unique_history:
            self.client.delete(self.history_queue)

    def available(self):
        """
        Returns:
            The current size of the queue.
        """
        return self.client.llen(self.queue)

    def push(self, element):
        """
        Append `element` to the queue.

        If `soft_capacity` is set and the queue is full, wait until a room is available.

        If `unique_history` is set, the ever-seen element is ignored.


        Args:
            element: an element of type that redis-py can encode.

        Returns:

        """
        if self.soft_capacity is not None:
            while self.available() > self.soft_capacity:
                if self.cooldown < 2: self.cooldown *= 2
                time.sleep(self.cooldown)
            if self.cooldown > 0.1:
                self.cooldown /= 2

        if self.unique_history:
            if not self.client.sismember(self.history_queue, element):
                self.client.rpush(self.queue, element)
                self.client.sadd(self.history_queue, element)
        else:
            self.client.rpush(self.queue, element)

    def pop_one(self):
        """
        Pop an element.
        Wait if the queue is empty.

        Returns:
            an element.
        """
        return self.client.blpop(keys=[self.queue])[1]

    def pop_at_least_one(self, max_num: int = 1):
        """
        Pop a batch of elements. return at least one.
        Wait if the queue is empty.
        Return immediately if the queue is not empty but smaller than `max_num`.

        Args:
            max_num (int): maximum number of samples to pop.

        Returns:
            a list of elements.
        """

        batch = [self.client.blpop(keys=[self.queue])[1]]

        for _ in range(1, max_num):
            element = self.client.lpop(self.queue)
            if element is None:
                break
            batch.append(element)

        return batch



import torch.utils.data
from collections import Counter, Iterable


def DataLoaderIterator(loader, num=None, last_step=0):
    """Convenient DataLoader wrapper when you need to iterate more than a full batch.

    It is recommended to set `drop_last=True` in your DataLoader.

    Example::

        loader = DataLoader(num_workers=8)
        for step, batch in DataLoaderIterator(loader):
            pass
        for step, _ in DataLoaderIterator(loader, num=6, last_step=2):
            print(step)  # 3, 4, 5, 6


    Args:
        loader (torch.utils.data.DataLoader)
        num (int, optional): Defaults to None. `None` means infinite iteration.
        last_step (int, optional): Defaults to 0. The iteration starts from `last_step+1`.
    """

    step = last_step
    while True:
        for batch in loader:
            step += 1
            if step > num:
                return
            yield step, batch


def BalancedDataLoader(dataset, classes, **kwargs):
    """If your dataset is unbalanced, this wrapper provides a uniform sampling.

    Example::

        # -3 is sampled twice as many as 2 or 3.
        loader = BalancedDataLoader([-3,5,2,3], ['R','G','B','B'], batch_size=1)

    Args:
        dataset (iterable): torch Dataset, list, or any sequence with known length.
        classes (iterable): A list of hashable type. Its length should be equal to the dataset.
        **kwargs: arguments to torch.utils.data.DataLoader
    """

    assert isinstance(classes, Iterable)
    assert len(dataset) == len(classes)
    counter = Counter(classes)
    for key in counter.keys():
        counter[key] = 1.0 / counter[key]
    weights = [counter[cls_] for cls_ in classes]

    kwargs['shuffle'] = False
    kwargs['batch_sampler'] = None
    kwargs['sampler'] = torch.utils.data.sampler.WeightedRandomSampler(
        weights=weights,
        num_samples=len(dataset),
        replacement=True
    )

    return torch.utils.data.DataLoader(dataset, **kwargs)



class _batch_known_len(object):
    def __init__(self, iterable, batch_size, total):
        self.iterable = iterable
        self.batch_size = batch_size
        self.total = total
        self.num_batch = ((total-1) // self.batch_size) + 1

    def __len__(self):
        return self.num_batch

    def __iter__(self):
        source = iter(self.iterable)

        for i in range(self.num_batch-1):
            chunk = [val for _, val in zip(range(self.batch_size), source)]
            yield chunk

        batch_size = self.total - (self.batch_size*(self.num_batch-1))
        chunk = [val for _, val in zip(range(batch_size), source)]
        yield chunk


class _batch_unknown_len(object):
    def __init__(self, iterable, batch_size):
        self.iterable = iterable
        self.batch_size = batch_size

    def __iter__(self):
        source = iter(self.iterable)

        while True:
            chunk = [val for _, val in zip(range(self.batch_size), source)]
            if not chunk:
                raise StopIteration
            yield chunk


def batch(iterable, batch_size, drop_last=False, total=None):
    """Bachify given iterable.

    if the given iterable has `__len__()`, this returns the iterable with `__len()__`.

    Example::

        import tqdm
        import tinder
        import time


        for i in tqdm.tqdm(tinder.batch(range(100), 4, total=14)):
            print(i)
            time.sleep(0.2)


        def f():
            for i in range(30):
                yield i


        for i in tqdm.tqdm(tinder.batch(f(), 4)):
            print(i)
            time.sleep(0.2)

    Args:
        iterable (iterable)
        batch_size (int)
        total (int, optional): if None, it tries to use `__len()__` of iterable.

    """
    if total == None and hasattr(iterable, '__len__'):
        total = len(iterable)

    if total == None:
        return _batch_unknown_len(iterable, batch_size)
    else:
        return _batch_known_len(iterable, batch_size, total)



import torch
import torch.nn as nn
from .nn import *
from .dataset import *
from .monitor import *
from .serving import *
from .runner import *
from .itertools import *


class AssertSize(nn.Module):
    """Assert the input has the specified size.

    Example::

        net = nn.Sequential(
            AssertSize(None, 3, 224, 224),
            nn.Conv2d(3, 64, kernel_size=1, stride=2),
            nn.Conv2d(64, 128, kernel_size=1, stride=2),
            AssertSize(None, 128, 64, 64),
        )

    Args:
        size (iterable): an iterable of dimensions. Each dimension is one of -1, None, or positive integer.

    """

    def __init__(self, *size):
        super().__init__()
        self.size = [s if s != -1 else None for s in size]

    def __repr__(self):
        return f'AssertSize({self.size})'

    def forward(self, x):
        size = x.size()
        if len(self.size) != len(size):
            raise RuntimeError(f"expected rank {len(self.size)} but got a tensor of rank {len(size)}")

        for expected, given in zip(self.size, size):
            if expected != None and expected != given:
                raise RuntimeError(f"expected size {self.size} but got a tensor of size {size}")


class Flatten(nn.Module):
    """A layer that flattens the input.

    Example::

        net = nn.Sequential(
            nn.Conv2d(..),
            nn.BatchNorm2d(..),
            nn.ReLU(),

            nn.Conv2d(..),
            nn.BatchNorm2d(..),
            nn.ReLU(),

            Flatten(),
            nn.Linear(3*3*512, 1024),
        )

    Args:
        x: input tensor
    """

    def forward(self, x):
        return x.view(x.size(0), -1)


class View(nn.Module):
    """nn.Module version of tensor.view().

    Example::

        layer = tinder.View(3, -1, 256)
        x = layer(x)

    The batch dimension is implicit.
    The above code is the same as `tensor.view(tensor.size(0), 3, -1, 256)`.

    Args:
        size_without_batch_dim (iterable): each dimension is one of -1, None, or positive.
    """

    def __init__(self, *size_without_batch_dim):
        super().__init__()
        self.size = [s if s != None else -1 for s in size_without_batch_dim]

    def forward(self, x):
        return x.view(x.size(0), *self.size)


def copy_opt_state(old: torch.optim.Optimizer, new: torch.optim.Optimizer):
    """Copy one optimizer's state to another.

    Example::


    Args:
        old (torch.optim.Optimizer): A source optimizer
        new (torch.optim.Optimizer): A destination optimizer
    """

    for group in new.param_groups:
        for parameter in group['params']:
            if parameter in old.state:
                # e.g. {'step', 'exp_avg', ..}
                new.state[parameter] = old.state[parameter]



def setup(*, trace=True, pdb_on_error=True, parse_args=True, logger=True, args=None):
    """
    Setup convinient utils to run a python script for deep learning.

    Examples:
          `./a.py train 4,5,6 lr=0.01`
          If a.py calls `tinder.setup()`, this is equivalent to `args=['train','4,5,6','lr=0.01]`.
          The following environments are set:

          - os.environ['mode'] = 'train'
          - os.environ['CUDA_VISIBLE_DEVICES'] = '4,5,6'
          - os.environ['CUDA_DEVICE_ORDER'] = 'PCI_BUS_ID'
          - os.environ['lr'] = '0.01'

    Args:
        trace: use `backtrace` module to print stacktrace
        pdb_on_error: enter `pdb` mode when an exception is raised
        parse_args: parse `args` or `sys.args` of a form `script.py [mode] [gpus_comma_separated]`
        logger: setup the logging format
        args (list): if given, this is used instead of `sys.args`
    """
    import os

    if trace:
        import backtrace
        backtrace.hook(align=True)
    if pdb_on_error:
        import sys
        old_hook = sys.excepthook

        def new_hook(type, value, tb):
            old_hook(type, value, tb)
            if type != KeyboardInterrupt:
                import pdb
                pdb.post_mortem(tb)

        sys.excepthook = new_hook

    os.environ['CUDA_DEVICE_ORDER'] = 'PCI_BUS_ID'

    if parse_args:
        import argparse
        parser = argparse.ArgumentParser()
        parser.add_argument('mode', help='cmd. e.g) train/val/test/preprocess', type=str)
        parser.add_argument('gpu', help='gpu list; e.g) 0,1', type=str)
        args, unknowns = parser.parse_known_args(args)

        os.environ['CUDA_VISIBLE_DEVICES'] = args.gpu
        os.environ['mode'] = args.mode
        for unknown in unknowns:
            if unknown[0] == '-': unknown = unknown[1:]
            if unknown[0] == '-': unknown = unknown[1:]
            idx = unknown.find('=')
            if idx == -1:
                os.environ[unknown] = '1'
                print('잘못된 입력: ', unknown)
                sys.exit(-1)
            else:
                os.environ[unknown[:idx]] = unknown[idx + 1:]

    if logger:
        import logging, tqdm
        log = logging.getLogger()
        log.setLevel(logging.DEBUG)
        formatter = logging.Formatter(
            '%(asctime)s [%(filename)16s:%(lineno)3s] %(message)s', datefmt='%m월%d일 %H시%M분')

        class TqdmLoggingHandler(logging.Handler):
            def emit(self, record):
                try:
                    msg = self.format(record)
                    tqdm.tqdm.write(msg)
                    self.flush()
                except (KeyboardInterrupt, SystemExit):
                    raise
                except Exception:
                    self.handleError(record)

        handler = TqdmLoggingHandler()
        handler.setFormatter(formatter)
        log.addHandler(handler)



import torch
import torch.nn as nn
from torch.autograd import Variable


class WeightScale(nn.Module):
    """A weight normalizing layer used in PGGAN.

    How it works:

        1. Initialize your ConvLayer with weights from N(0, std=kaiming)
        2. WeightScale calculates the initial std of weights (from 1)
        3. Divide Conv.weights by std (from 2). Now Conv.weights are ~N(0,1)
        4. On forward, WeightScale multiply the input by std (from 2).
        5. Bias or Activation is applied after WeightScale. If Conv has bias, WeightScale steals it.

    Note that the scale factor calculated in 2 is constant permanently.

    Advantage:

        The author of PGGAN claims Adam is better at training weights in N(0,1) than training conv weights of different std.

    Example::

        conv = nn.Conv2d(3, 64, kernel_size=3, padding=1)
        ws = WeightScale(conv)
        nn.Sequential(
            conv,
            ws
        )


    Args:
        prev_layer: a layer (e.g. nn.Conv2d) with `weight` and optionally `bias`.
        init_with_leakiness: if given, it initializes `prev_layer` with `kaiming_normal`.
    """

    def __init__(self, prev_layer, init_with_leakiness: float = None):
        super().__init__()

        if init_with_leakiness is not None:
            torch.nn.init.kaiming_normal(prev_layer.weight, a=init_with_leakiness)

        self.scale = torch.nn.Parameter(torch.cuda.FloatTensor(1), requires_grad=False)
        self.scale.data[0] = torch.mean(prev_layer.weight.data ** 2) ** 0.5 + 1e-8
        prev_layer.weight.data.div_(self.scale.data[0])

        if prev_layer.bias is not None:
            self.bias = prev_layer.bias
            self.bias.data.zero_()
            prev_layer.bias = None
        else:
            self.bias = None

    def forward(self, x):
        x = self.scale * x
        if self.bias is not None:
            x += self.bias.view(1, self.bias.size(0), 1, 1)
        return x

    def __repr__(self):
        # param_str = '(prev_layer = %s)' % (self.prev_layer.__class__.__name__)
        return self.__class__.__name__


class PixelwiseNormalize(nn.Module):
    """Pixelwise Normalization used in PGGAN.
        It normalizes the input [B,C,H,W] so that the L2 norms over the C dimension is 1.
        There are B*H*W norms.

        Example::

            x = tinder.PixelwiseNormalize()(x)
    """

    def forward(self, x):
        denominator = (x.pow(2).mean(dim=1, keepdim=True) + 1e-8).sqrt()
        return x / denominator


# calculate std deviation for each feature in each location
# scalar = average over all feature & locations
# broadcast to constant feature map
class MinibatchStddev(nn.Module):
    """Layer for GAN Discriminator used in PGGAN.
        It penalizes when images in the minibatch look similar.
        For example, G generates fall into mode collapse and generate similar images,
        then stddev of the minibatch is small. D looks at the small stddev and thinks this is likely fake.

        This layer calculates stddev and provide it as additional channel.
    """

    def __init__(self):
        super().__init__()

    def forward(self, x):
        """
        Args:
            x (N, C, H, W)

        Returns:
            (N, C+1, H, W): The last channel dimension is stddev.
        """

        std = ((x - x.mean(dim=0, keepdim=True)).pow(2).mean(dim=0, keepdim=True) + 1e-8).sqrt().mean()
        scalar = std.expand(x.size(0), 1, x.size(2), x.size(3))  # [N,1,H,W]
        x = torch.cat([x, scalar], dim=1)
        return x


def loss_wgan_gp(D, real: torch.cuda.FloatTensor, fake: torch.cuda.FloatTensor) -> torch.autograd.Variable:
    """Gradient Penalty for Wasserstein GAN.

    It generates random interpolations between the real and fake points,
    and tries to make grad norm close to 1.

    Note:
        - The second-order derivative is unstable in PyTorch.
          Some layers in your G/D may not work.
        - It assumes CUDA.

    Example::

        gp = tinder.wgan_gp(D, real_img, fake_img)
        loss = lambda*gp

    Args:
        D (callable): A discriminator. Typically `nn.Module` or a lambda function that returns the score.
        real (torch.cuda.FloatTensor): real sample. Note this is not a `Variable`.
        fake (torch.cuda.FloatTensor): fake sample. Note this is not a `Variable`.

    Returns:
        (Variable): gradient penalty.
    """

    batch_size = real.size(0)
    e = torch.cuda.FloatTensor(batch_size).uniform_().view(batch_size, 1, 1, 1)
    x_hat = Variable(e * real + (1 - e) * fake, requires_grad=True)
    scores = D(x_hat)
    grads = torch.autograd.grad(scores, x_hat, retain_graph=True, create_graph=True,
                                grad_outputs=torch.ones_like(scores)
                                )[0]

    norms = grads.view(batch_size, -1).norm(2, dim=1)
    return ((norms - 1) ** 2).mean()



from types import SimpleNamespace


class Stat(object):
    """A small class that calculates the statistics for a scalar metric.
    This is useful to calcaulate accuracies and losses across minibatches.

    Args:
        alpha (float): Defaults to 1. Weight(0~1) used for exponential moving average.
            A smaller weight reprsents more smooth average.

    Attributes:
        count (int): the number of samples.
        sum (float)
        average (float): the average of samples. read-only.
        ema (float): exponential moving average.
        alpha (float): coefficient for exponential moving average.
    """

    def __init__(self, alpha: float = 0.6):
        self.alpha = alpha
        self.clear()

    def update(self, value, count: int = 1):
        """Update stats with new samples.

        Args:
            value (int or float): new sample, or average of multiple samples.
            count (int): Defaults to 1. This is useful when batch size is not uniform.
        """

        self.count += count
        self.sum += value

        value /= count
        if self.ema is None:
            self.ema = value
        else:
            self.ema = self.alpha * value + (1 - self.alpha) * self.ema

        return self.ema

    @property
    def average(self) -> float:
        """Average of all samples.

        Returns:
            average
        """

        if self.count > 0:
            return self.sum / self.count
        return None

    def clear(self):
        """Clear the history and reset stats to zero.
        """

        self.ema = None  # ema = Exponential Moving Average
        self.count = 0
        self.sum = 0


class Stats(SimpleNamespace):
    """A class that holds named stats.
    """

    def __init__(self, **stats):
        super().__init__(**stats)

    def clear(self):
        """Clear all stats.
        """

        for stat in self.__dict__.values():
            stat.clear()



from setuptools import setup

LONG_DESCRIPTION = """
**tinder** provides extra layers and helpers for Pytorch.
"""

setup(name='tinder',
      version='0.1',
      description='Pytorch helpers and utils',
      long_description=LONG_DESCRIPTION,
      classifiers=[
          'Operating System :: OS Independent',
          'Intended Audience :: Science/Research',
          'Programming Language :: Python :: 3.6',
          'Topic :: Scientific/Engineering'
      ],
      url='http://github.com/elbaro/tinder',
      author='elbaro',
      author_email='elbaro@users.noreply.github.com',
      license='MIT',
      packages=['tinder'],
      keywords=['tinder','pytorch','torch'],
      zip_safe=False)

